/*
 * Wireless2.4GHz.cpp
 *
 *  Created on: Mar 4, 2011
 *      Author: Andygongyb
 */

#include "Wireless2.4GHz.hpp"

using namespace wireless_names;

CWireless *CWireless::m_instance = NULL;

CWireless::CWireless(void) {
	CSPI::init(CSPI::D2);
	PORTD |= _BV(PD1);
	DDRD &= ~_BV(PD1);

	m_tx_head = m_data_bank + 20; // we need dummy queue heads
	m_tx_tail = m_tx_head;
	m_tx_len = 0;

	m_rx_head = m_data_bank + 21;
	m_rx_tail = m_rx_head;
	m_rx_len = 0;

	m_failure_count = 0;
	m_free_count = 0;

	m_free_block = m_data_bank + 22;

	for (uint8_t i = 0; i < 20; i++) {
		CData *ptr = m_data_bank + i;
		add_to_free_list(ptr);
	}

	io_set csn = { &DDRB, &PORTB, &PINB, PB7 };

	CSPIDevice::init(&csn);

	m_pin_ce = _BV(PD0);
	DDRD |= m_pin_ce;
	m_port_ce = &PORTD;
	*m_port_ce &= ~m_pin_ce;

	ce_down();

	portENTER_CRITICAL();
	write_reg(CONFIG, _BV(EN_CRC) | _BV(CRCO)); // power down
	write_reg(CONFIG, _BV(PWR_UP) | _BV(PRIM_RX) | _BV(EN_CRC) | _BV(CRCO)); // power up in PRX mode
	cmd(FLUSH_TX);
	cmd(FLUSH_RX);
	write_reg(STATUS, 0x70); // clear the interrupt
	write_reg(RF_SETUP, 0x0E); // set the data rate to 2Mbps with 0dBm output power
	write_reg(EN_AA, 0x3F); // enable auto acknowledgment for 6 pipes
	write_reg(EN_RXADDR, 0x01); // enable pipe 0
	write_reg(DYNPD, 0x3F); // enable dynamic payload length in 6 pipes
	write_reg(FEATURE, _BV(EN_DPL) | _BV(EN_ACK_PAY)); // enable global dynamic payload length and payload with ACK
	write_reg(SETUP_AW, 0x03); // set address width to 5
	for (uint8_t i = 0; i < 6; i++)
		write_reg(RX_PW_P0 + i, 0x20); // set payload length to 32 bytes
	_delay_ms(1.3); // wait for power up

	property_t property = { 50, { 0xE7, 0xE7, 0xE7, 0xE7, 0xC1 }, { 0xE7, 0xE7,
			0xE7, 0xE7 }, { 0xC2, 0xC3, 0xC4, 0xC5, 0xC6 }, 10, 10 };

	set_property(&property);

	EIMSK |= _BV(INT1);
	EICRA |= _BV(ISC11); // external interrupt, falling edge trigger
	portEXIT_CRITICAL()
	;
}

CWireless::~CWireless(void) {
	ce_down(); // standby
	write_reg(CONFIG, _BV(EN_CRC) | _BV(CRCO)); // turn off the module
}

uint8_t CWireless::append(uint8_t *buffer, uint8_t len, bool *if_done,
		uint8_t pipe_id) {
	portENTER_CRITICAL();
	if (m_tx_len == m_property.max_tx_len) {
		if (m_tx_head->m_next->m_if_done)
			*m_tx_head->m_next->m_if_done = true;
		m_tx_len--;
		CData *ptr = m_tx_head;
		m_tx_head = ptr->m_next;
		add_to_free_list(ptr);
	}
	m_tx_tail->m_next = get_from_free_list();
	m_tx_tail = m_tx_tail->m_next;
	if (!m_tx_len) {
		ce_down();
		cmd(W_ACK_PAYLOAD | (0x07 & pipe_id), buffer, len); // if the ACK payload buffer is empty, then fill it
		ce_up();
	} else {
		memcpy(m_tx_tail->m_buffer, buffer, len);
	}
	m_tx_tail->init(len, if_done, pipe_id);

	m_tx_len++;
	portEXIT_CRITICAL();
	return m_tx_len; // return the number of data blocks currently in the queue
}

uint8_t CWireless::get(uint8_t* buffer) {
	if (m_rx_len) {
		portENTER_CRITICAL();

		memcpy(buffer, m_rx_head->m_next->m_buffer, m_rx_head->m_next->m_len);
		uint8_t ret = m_rx_head->m_next->m_len;

		CData* ptr = m_rx_head;
		m_rx_head = ptr->m_next;
		m_rx_len--;

		add_to_free_list(ptr);

		portEXIT_CRITICAL();
		return ret; // return the length of the data block
	}
	return 0;
}

uint8_t CWireless::next_len(void) {
	if (m_rx_len) {
		return m_rx_head->m_next->m_len;
	}
	return 0;
}

void CWireless::set_property(property_t *property) {
	portENTER_CRITICAL();
	ce_down(); // standby

	write_reg(RF_CH, (property->channel) & 0x7F); // set the channel
	uint8_t buffer[5];
	buffer[0] = property->rx_addr_lsb[0];
	for (uint8_t i = 4; i; i--) {
		*(buffer + i) = property->rx_addr_domain[4 - i]; // reverse the address domain
		write_reg(RX_ADDR_P1 + i, property->rx_addr_lsb[i]); // set the LSByte of pipe2 ~ pipe5 addresses
	}
	write_reg(RX_ADDR_P1, buffer, 5); // set the domain together with the pipe1 LSByte
	for (uint8_t i = 5; i; i--) {
		*(buffer + i - 1) = property->rx_p0_addr[5 - i]; // reverse the pipe0 address
		write_reg(RX_ADDR_P0, buffer, 5); // set the pipe0 address
	}

	ce_up(); // resume to PRX mode
	memcpy((void*) &m_property, (const void*) property, sizeof(property_t));
	portEXIT_CRITICAL()
	;
}

uint8_t CWireless::read_reg(uint8_t addr, uint8_t *data) {
	portENTER_CRITICAL();
	start();
	uint8_t ret = CSPI::trans_char(R_REGISTER | (REGISTER_MASK & addr));
	*data = CSPI::trans_char(NOP);
	stop();
	portEXIT_CRITICAL();
	return ret;
}

uint8_t CWireless::write_reg(uint8_t addr, uint8_t data) {
	portENTER_CRITICAL();
	start();
	uint8_t ret = CSPI::trans_char(W_REGISTER | (REGISTER_MASK & addr));
	CSPI::trans_char(data);
	stop();
	portEXIT_CRITICAL();
	return ret;
}

uint8_t CWireless::cmd(uint8_t cmd, uint8_t *send, uint8_t len,
		uint8_t *receive) {
	start();
	uint8_t ret = CSPI::trans_char(cmd);
	CSPI::trans_block(send, len, receive);
	stop();
	return ret;
}

void INT1_vect(void) {

	static CWireless *ptr = CWireless::instance();

	uint8_t status = ptr->cmd(NOP); // read the status register
	ptr->write_reg(CWireless::STATUS, 0x7E);
	ptr->ce_down();
	if (status & _BV(TX_DS)) {
		if (ptr->m_tx_head->m_next->m_if_done)
			*ptr->m_tx_head->m_next->m_if_done = true;
		ptr->m_tx_len--;
		CWireless::CData *ptr2 = ptr->m_tx_head;
		ptr->m_tx_head = ptr2->m_next;
		ptr->add_to_free_list(ptr2);

		if (ptr->m_tx_len) {
			
			ptr->cmd(
					W_ACK_PAYLOAD | (0x07 & ptr->m_tx_head->m_next->m_pipe_id),
					ptr->m_tx_head->m_next->m_buffer,
					ptr->m_tx_head->m_next->m_len);
		}
	}

	if (status & _BV(RX_DR)) {
		if (ptr->m_rx_len == ptr->m_property.max_rx_len) {
			ptr->m_rx_len--;
			CWireless::CData *ptr2 = ptr->m_rx_head;
			ptr->m_rx_head = ptr2->m_next;

			ptr->add_to_free_list(ptr2);
		}
		uint8_t pipe_id = (status & RX_P_NO) >> 1, len = NOP;
		ptr->cmd(R_RX_PL_WID, &len, 1, &len);
		ptr->m_rx_tail->m_next = ptr->get_from_free_list();
		if (ptr->m_rx_tail->m_next == NULL) {
			while (1)
				;
		}
		ptr->m_rx_tail = ptr->m_rx_tail->m_next;
		ptr->m_rx_tail->init(len, NULL, pipe_id);
		ptr->cmd(R_RX_PAYLOAD, ptr->m_rx_tail->m_buffer, len,
				ptr->m_rx_tail->m_buffer);
		ptr->m_rx_len++;
	}
	ptr->ce_up();
}

